@@ -538,8 +538,8 @@ bool printf_printf(const struct format_T *format, const wchar_t *arg)
 {
     switch (format->type) {
 	case ft_none:
-	    assert(format->convspec[0] == '%');
-	    printf(format->convspec);
+	    assert(strcmp(format->convspec, "%%") == 0);
+	    putchar('%');
 	    return false;
 	case ft_raw:
 	    assert(format->convspec[0] != '%');
