@@ -795,14 +795,14 @@ redir_fail:
 pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 {
     sigset_t all, savemask;
-    if (sigtype & (t_quitint | t_tstp)) {
+    bool sigblock = is_interactive_now || doing_job_control_now
+	|| (sigtype & (t_quitint | t_tstp));
+    if (sigblock) {
 	sigfillset(&all);
 	sigemptyset(&savemask);
 	if (sigprocmask(SIG_BLOCK, &all, &savemask) < 0)
-	    xerror(errno, "sigprocmask(BLOCK, all)");
+	    xerror(errno, "sigprocmask");
     }
-    restore_job_signals();
-    restore_interactive_signals();
 
     pid_t cpid = fork();
 
@@ -815,7 +815,6 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 	    if (doing_job_control_now && pgid >= 0)
 		setpgid(cpid, pgid);
 	}
-	set_signals();
     } else {
 	/* child process */
 	bool save_is_interactive_now = is_interactive_now;
@@ -827,6 +826,8 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 	}
 	forget_initial_pgid();
 	neglect_all_jobs();
+	restore_job_signals();
+	restore_interactive_signals();
 	if (sigtype & t_leave) {
 	    clear_shellfds(false);
 	} else {
@@ -841,9 +842,9 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 	    if (save_is_interactive_now)
 		ignore_sigtstp();
     }
-    if (sigtype & (t_quitint | t_tstp))
+    if (sigblock)
 	if (sigprocmask(SIG_SETMASK, &savemask, NULL) < 0 && errno != EINTR)
-	    xerror(errno, "sigprocmask(SETMASK, none)");
+	    xerror(errno, "sigprocmask");
     return cpid;
 }
 
@@ -1847,7 +1848,7 @@ bool print_command_info_human_friendlily(const wchar_t *commandname)
 	goto done;
     }
 
-    xerror(0, Ngt("%s: not found"), name);
+    xerror(0, Ngt("%s: no such command or function"), name);
     free(name);
     return false;
 done:
@@ -1864,6 +1865,8 @@ bool print_command_fullpath(const wchar_t *commandname, bool hf)
     if (!name)
 	return false;
     if (!is_executable_regular(name)) {
+	if (hf)
+	    xerror(0, Ngt("%s: no such command or function"), name);
 	free(name);
 	return false;
     }
