@@ -159,7 +159,7 @@ bool is_readable_regular(const char *path)
 /* Checks if `path' is an executable regular file. */
 bool is_executable_regular(const char *path)
 {
-    return is_regular_file(path) && access(path, X_OK) == 0;
+    return is_regular_file(path) && is_executable(path);
 }
 
 /* Checks if `path' is a directory. */
